from pathlib import Path
from typing import Dict, Optional, Union

import pandas as pd

from readii.io.loaders.general import loadFileToDataFrame
from readii.utils import logger


def loadFeatureFilesFromImageTypes(extracted_feature_dir:Union[Path|str], # noqa
                                   image_types:list, 
                                   drop_labels:Optional[bool]=True, 
                                   labels_to_drop:Optional[list]=None)->Dict[str,pd.DataFrame]: 
    """Load in all the specified extracted imaging feature sets from a directory and return them as a dictionary of dataframes.

    Parameters
    ----------
    extracted_feature_dir : str
        Path to the directory containing the extracted feature csv files
    image_types : list, optional
        List of image types to load in. The default is ['original'].
    drop_labels : bool, optional
        Whether to drop the labels from the dataframes. Use when loading labelled data from data_setup_for_modeling.ipynb. The default is True.
    labels_to_drop : list, optional
        List of labels to drop from the dataframes. The default is ["patient_ID","survival_time_in_years","survival_event_binary"] based on code
        in data_setup_for_modeling.ipynb.

    Returns
    -------
    feature_sets : dict
        Dictionary of dataframes containing the extracted radiomics features.
    """
    # Convert directory to Path object if it is a string
    if isinstance(extracted_feature_dir, str):
        extracted_feature_dir = Path(extracted_feature_dir)

    # Set default labels to drop if not specified
    if labels_to_drop is None:
        labels_to_drop = ["patient_ID","survival_time_in_years","survival_event_binary"]

    # Initialize dictionary to store the feature sets
    feature_sets = {}

    # Check if the passed in extracted feature directory exists
    if not extracted_feature_dir.exists() or not extracted_feature_dir.is_dir():
        msg = f"Extracted feature directory {extracted_feature_dir} does not exist."
        logger.error(f"Extracted feature directory {extracted_feature_dir} does not exist.")
        raise FileNotFoundError()
    
    # Get list of all the csv files in the directory with their full paths
    feature_file_list = sorted(extracted_feature_dir.glob("*.csv"))

    # Loop through all the files in the directory
    for image_type in image_types:
        try:
            # Extract the image type feature csv file from the feature directory  
            matching_files = [file for file in feature_file_list if (image_type in file.stem)]  

            match len(matching_files):
                case 1:
                    # Only one file found, use it  
                    pass
                case 0:
                    # No files found for this image type
                    logger.warning(f"No {image_type} feature csv files found in {extracted_feature_dir}")
                    # Skip to the next image type
                    continue
                case _:
                    # Multiple files found
                    msg = f"Multiple {image_type} feature csv files found in {extracted_feature_dir}. First one will be used."
                    logger.warning(msg)

            feature_file_path = matching_files[0]  
            # Remove the image type file from the list of feature files  
            feature_file_list.remove(feature_file_path)

        except Exception as e:
            logger.warning(f"Error loading {image_type} feature csv files from {extracted_feature_dir}: {e}")
            raise e
            
        # Load the feature data into a pandas dataframe
        raw_feature_data = loadFileToDataFrame(feature_file_path)

        try:
            # Drop the labels from the dataframe if specified
            if drop_labels:
                # Data is now only extracted features
                raw_feature_data.drop(labels_to_drop, axis=1, inplace=True)

        except KeyError:
            logger.warning(f"{feature_file_path} does not have the labels {labels_to_drop} to drop.")
            # Skip to the next image type
            continue

        # Save the dataframe to the feature_sets dictionary
        feature_sets[image_type] = raw_feature_data
    # end image type loop

    # After processing all image types, check if any feature sets were loaded 
    if not feature_sets:
        logger.error(f"No valid feature sets were loaded from {extracted_feature_dir}")
        raise ValueError()
    
    return feature_sets


from typing import Optional

from pandas import DataFrame

from readii.utils import logger


def replaceColumnValues(dataframe:DataFrame,
                       column_to_change:str,
                       replacement_value_data:dict
                       ) -> DataFrame:
    """Replace specified values in a column with a new value.

    Parameters
    ----------
    dataframe : DataFrame
        Dataframe to replace values in.
    column_to_change : str
        Name of the column to replace values in.
    replacement_value_data : dict
        Dictionary of values to replace in the column. Key is the new value, value is the old value(s) to replace.
        Can be a single value or a list of values.
    
    Returns
    -------
    dataframe : DataFrame
        Dataframe with values replaced.
    """
    # Check if the column name is a valid column in the dataframe
    if column_to_change not in dataframe.columns:
        msg = f"Column {column_to_change} not found in dataframe."
        logger.exception(msg)
        raise ValueError(msg)
    
    for new_value, old_values in replacement_value_data.items():
        # Check if the replacement value is a valid value in the column
        values_not_found_in_column = set(old_values).difference(set(dataframe[column_to_change].unique()))

        if values_not_found_in_column == set(old_values):
            msg = f"All values in {values_not_found_in_column} are not found to be replaced in column {column_to_change}."
            logger.exception(msg)
            raise ValueError(msg)
        
        # Replace the old values with the new value
        dataframe = dataframe.replace(to_replace=old_values, 
                                      value=new_value)
        
    return dataframe


def splitDataByColumnValue(dataframe:DataFrame,
                           split_col_data:dict[str,list],
                           impute_value:Optional[object | None] = None,
                           ) -> dict[str,DataFrame]:
    """Split a dataframe into multiple dataframes based on the values in a specified column. Optionally, impute values in the split columns.

    Parameters
    ----------
    dataframe : DataFrame
        Dataframe to split.
    split_col_data : dict[str,list]
        Dictionary of a column name and values to split the dataframe by. Key is the column name, value is the list of values to split by.
    impute_value (Optional)
        If set, will impute any non-specified split values in the split column with this value. The default is None.
    
    Returns
    -------
    split_dataframes : dict
        Dictionary of dataframes, where the key is the split value and the value is the dataframe for that split value.
    """
    # Initialize dictionary to store the split dataframes
    split_dataframes = {}

    for split_column_name, split_col_values in split_col_data.items():
        # Check if the column name is a valid column in the dataframe
        if split_column_name not in dataframe.columns:
            msg = f"Column {split_column_name} not found in dataframe."
            logger.exception(msg)
            raise ValueError(msg)
        
        if impute_value is not None:
            # Get all values in the column that are not one of the split_col_values
            column_value_set = set(dataframe[split_column_name].unique())
            split_value_set = set(split_col_values)
            non_split_values = list(column_value_set - split_value_set)

            # Replace all values not specified in the split_col_data with the impute_value specified for this column
            dataframe = replaceColumnValues(dataframe,
                                            column_to_change=split_column_name,
                                            replacement_value_data={impute_value: non_split_values})
            
        # End imputation
        
        # Split dataframe by the specified split_col_values
        for split_value in split_col_values:
            # Check if the split_value is a valid value in the column
            if split_value not in dataframe[split_column_name].unique():
                msg = f"Split value {split_value} not found in column {split_column_name}."
                logger.exception(msg)
                raise ValueError(msg)
            
            # Split the dataframe by the specified split_value
            split_dataframe = dataframe[dataframe[split_column_name] == split_value]

            split_dataframe.reset_index(inplace=True, drop=True)

            # Save the split dataframe to the dictionary
            split_dataframes[split_value] = split_dataframe

        # End dataframe splitting
    
    return split_dataframes



import SimpleITK as sitk
from SimpleITK import Image
import numpy as np

from readii.image_processing import alignImages

from typing import Optional, Union
from numpy import ndarray

def getArrayFromImageOrArray(imageOrArray: Union[Image, ndarray]) -> ndarray:
    """Function to convert a SimpleITK Image to a numpy array.
    
    Parameters
    ----------
    imageOrArray : sitk.Image | np.ndarray
        Image or array to convert to numpy array.

    Returns
    -------
    np.ndarray
        Numpy array version of the input image or array.
        
    Raises
    ------
    ValueError
        If the input is not a SimpleITK Image or numpy array.
    """
    assert isinstance(imageOrArray, Image) or isinstance(imageOrArray, ndarray), \
        "Input must be a SimpleITK Image or numpy array."

    if isinstance(imageOrArray, Image):
        return sitk.GetArrayFromImage(imageOrArray)
    elif isinstance(imageOrArray, ndarray):
        return imageOrArray    

def makeShuffleImage(
    baseImage: Union[Image, ndarray],
    randomSeed: Optional[int] = None,
) -> Union[Image, ndarray]:
    
    """Function to shuffle all pixel values in a sitk.Image or np.ndarray (developed for 3D, should work on 2D as well)

    Parameters
    ----------
    baseImage : sitk.Image | np.ndarray
        Image to shuffle the pixels in. Can be a sitk.Image or np.ndarray.
    randomSeed : int
        Value to initialize random number generator with. Set for reproducible results.
        
    Returns
    -------
    sitk.Image | np.ndarray
        Image with all pixel values randomly shuffled with same dimensions and object type as input image.
    """
    # # Check if baseImage is a sitk.Image or np.ndarray
    arrImage = getArrayFromImageOrArray(baseImage)

    # Get array dimensions to reshape back to
    imgDimensions = arrImage.shape

    # Flatten the 3D array to 1D so values can be shuffled
    flatArrImage = arrImage.flatten()

    # Set the random seed for np random generator
    randNumGen = np.random.default_rng(seed=randomSeed)

    # Shuffle the flat array
    randNumGen.shuffle(flatArrImage)

    # Reshape the array back into the original image dimensions
    shuffled3DArrImage = np.reshape(flatArrImage, imgDimensions)

    if type(baseImage) == sitk.Image:
        # Convert back to sitk Image
        shuffledImage = sitk.GetImageFromArray(shuffled3DArrImage)

        # Set the origin/direction/spacing from original image to shuffled image
        alignedShuffledImage = alignImages(baseImage, shuffledImage)

        # Return the shuffled image
        return alignedShuffledImage
    
    else:
        # Return the shuffled array
        return shuffled3DArrImage

def makeRandomImage(
    baseImage: Union[Image, ndarray],
    randomSeed: Optional[int] = None,
) -> Union[sitk.Image, np.ndarray]:
    """Function to generate random pixel values based on the range of values in a sitk.Image or np.ndarray (developed for 3D, should work on 2D as well)

    Parameters
    ----------
    baseImage : sitk.Image | np.ndarray
        Image to randomly generate pixel values. Can be a sitk.Image or np.ndarray.
    randomSeed : int
        Value to initialize random number generator with. Set for reproducible results.

    Returns
    -------
    sitk.Image | np.ndarray
        Image with all pixel values randomly generated with same dimensions and object type as input image.
    """
    # # Check if baseImage is a sitk.Image or np.ndarray
    arrImage = getArrayFromImageOrArray(baseImage)

    # Get array dimensions to reshape back to
    imgDimensions = arrImage.shape

    # Get min and max HU values to set as range for random values
    minVoxelVal = np.min(arrImage)
    maxVoxelVal = np.max(arrImage)

    # Delete arrImage to save memory
    del arrImage

    # Set the random seed for np random generator
    randNumGen = np.random.default_rng(seed=randomSeed)

    # Generate random array with same dimensions as baseImage with values ranging from the minimum to maximum inclusive of the original image
    random3DArr = randNumGen.integers(
        low=minVoxelVal, high=maxVoxelVal, endpoint=True, size=imgDimensions
    )

    if type(baseImage) == sitk.Image:
        # Convert random array to a sitk Image
        randomImage = sitk.GetImageFromArray(random3DArr)

        # Set the origin/direction/spacing from the original image to the random image
        alignedRandomImage = alignImages(baseImage, randomImage)

        # Return the random image
        return alignedRandomImage
    
    else:
        # Return the random array
        return random3DArr
    
def makeRandomSampleFromDistributionImage(
    baseImage: Union[Image, ndarray],
    randomSeed: Optional[int] = None,
) -> Union[sitk.Image, np.ndarray]:
    """Function to randomly sample all the pixel values the distribution of existing values in a sitk.Image or np.ndarray.

    Parameters
    ----------
    imageToRandomize : sitk.Image | np.ndarray
        Image to randomly sample the pixels from. Can be a sitk.Image or np.ndarray.
    randomSeed : int
        Value to initialize random number generator with. Set for reproducible results.
    Returns
    -------
    sitk.Image | np.ndarray
        Image with all pixel values randomly sampled from the initial dstribution of the image, with same dimensions and object type as input image.
    """
    # Check if baseImage is a sitk.Image or np.ndarray
    arrImage = getArrayFromImageOrArray(baseImage)

    # Get array dimensions to reshape back to
    imgDimensions = arrImage.shape

    # Flatten the 3D array to 1D so values can be shuffled
    flatArrImage = arrImage.flatten()

    # Set the random seed for np random number generator
    randNumGen = np.random.default_rng(seed=randomSeed)

    # Randomly sample values for new array from original image distribution
    sampled_array = randNumGen.choice(flatArrImage, size=len(flatArrImage), replace=True)

    # Reshape the array back into the original image dimensions
    randomlySampled3DArrImage = np.reshape(sampled_array, imgDimensions)

    if type(baseImage) == sitk.Image:
        # Convert back to sitk Image
        randomlySampledImage = sitk.GetImageFromArray(randomlySampled3DArrImage)

        # Set the origin/direction/spacing from original image to sampled image
        alignedRandomlySampledImage = alignImages(baseImage, randomlySampledImage)
        
        # Return the randomly sampled image 
        return alignedRandomlySampledImage
    else:
        # Return the randomly sampled array
        return randomlySampled3DArrImage

def negativeControlROIOnly(
        baseImage: Union[Image, ndarray], 
        roiMask: Union[Image, ndarray], 
        negativeControlType: str = "shuffled",
        randomSeed: Optional[int] = None
        ) -> Union[Image, ndarray]:
    """Function to apply a negative control to a ROI only, without changing the background of the image.

    Parameters
    ----------
    baseImage : sitk.Image | np.ndarray
        Image to apply negative control to. Can be a sitk.Image or np.ndarray.
    roiMask : sitk.Image | np.ndarray
        Mask of the ROI to apply negative control within. Can be a sitk.Image or np.ndarray.
    negativeControlType : {'shuffled', 'randomized', 'randomized_sampled'}, default 'shuffled'
        Name of negative control to apply.
    randomSeed : int, default None    
        Value to initialize random number generator with. Set for reproducible results.
    
    Returns
    -------
    sitk.Image | np.ndarray
        Image with negative control function applied to all pixel values within the ROI.
    """

    if negativeControlType not in ["shuffled", "randomized", "randomized_sampled"]:
        raise ValueError("negativeControlType must be one of 'shuffled', 'randomized', or 'randomized_sampled'")
    
    # Check if baseImage is a sitk.Image or np.ndarray
    arrBaseImage = getArrayFromImageOrArray(baseImage)
    
    # Check if roiMask is a sitk.Image or np.ndarray
    arrROIMask = getArrayFromImageOrArray(roiMask)

    # Get binary segmentation masks
    # ROI is 1, background is 0
    binROIMask = np.where(arrROIMask > 0, 1, 0)
    if binROIMask.any() == False:
        raise ValueError("ROI mask is all 0s. No pixels in ROI to apply negative control to. ROI pixels should be > 1.")

    # Get just ROI pixels
    maskIndices = np.nonzero(binROIMask)
    # Get a 1D array of just the ROI pixels
    flatROIBaseValues = arrBaseImage[maskIndices]

    # Get desired negative control of baseImage
    arrNCROIValues = applyNegativeControl(baseImage = flatROIBaseValues,
                                          negativeControlType = negativeControlType,
                                          negativeControlRegion = "full",
                                          randomSeed = randomSeed)

    arrBaseImage[maskIndices] = arrNCROIValues

    # # Apply negative control to ROI pixels and keep original non-ROI pixels
    # arrNCROIImage = (arrNCBaseImage * binROIMask) + (arrBaseImage * inverseBinROIMask)

    if type(baseImage) == sitk.Image:
        # Convert back to sitk Image
        ncROIImage = sitk.GetImageFromArray(arrBaseImage)
        
        # Set the origin/direction/spacing from original image to negative control image
        alignedNCROIImage = alignImages(baseImage, ncROIImage)
        
        # Return the negative control image
        return alignedNCROIImage
    else:
        # Return the negative control array
        return arrBaseImage


def negativeControlNonROIOnly(
        baseImage: Union[Image, ndarray], 
        roiMask: Union[Image, ndarray], 
        negativeControlType: str = "shuffled",
        randomSeed: Optional[int] = None
        ) -> Union[Image, ndarray]:
    """Function to apply a negative control to all pixel values outside the ROI, without changing the ROI pixels.

    Parameters
    ----------
    baseImage : sitk.Image | np.ndarray
        Image to apply negative control to. Can be a sitk.Image or np.ndarray.
    roiMask : sitk.Image | np.ndarray
        Mask of the ROI to keep original image values within. Can be a sitk.Image or np.ndarray.
    negativeControlType : {'shuffled', 'randomized', 'randomized_sampled'}, default 'shuffled'
        Name of negative control to apply.
    randomSeed : int, default None    
        Value to initialize random number generator with. Set for reproducible results.
    
    Returns
    -------
    sitk.Image | np.ndarray
        Image with negative control function applied to all pixel values within the ROI.
    """

    if negativeControlType not in ["shuffled", "randomized", "randomized_sampled"]:
        raise ValueError("negativeControlType must be one of 'shuffled', 'randomized', or 'randomized_sampled'")
    
    # Check if baseImage is a sitk.Image or np.ndarray
    arrBaseImage = getArrayFromImageOrArray(baseImage)

    # Check if roiMask is a sitk.Image or np.ndarray
    arrROIMask = getArrayFromImageOrArray(roiMask)

    # Get binary segmentation masks
    # ROI is 1, background is 0
    binNonROIMask = np.where(arrROIMask > 0, 0, 1)
    if binNonROIMask.any() == False:
        raise ValueError("ROI mask is all 0s. No pixels in ROI to apply negative control to. ROI pixels should be > 1.")

    # Get just ROI pixels
    maskIndices = np.nonzero(binNonROIMask)
    # Get a 1D array of just the ROI pixels
    flatNonROIBaseValues = arrBaseImage[maskIndices]

    # Get desired negative control of baseImage
    arrNCNonROIValues = applyNegativeControl(baseImage = flatNonROIBaseValues,
                                          negativeControlType = negativeControlType,
                                          negativeControlRegion = "full",
                                          randomSeed = randomSeed)
    
    arrBaseImage[maskIndices] = arrNCNonROIValues

    if type(baseImage) == sitk.Image:
        # Convert back to sitk Image
        ncNonROIImage = sitk.GetImageFromArray(arrBaseImage)
        
        # Set the origin/direction/spacing from original image to negative control image
        alignedNCNonROIImage = alignImages(baseImage, ncNonROIImage)
        
        # Return the negative control image
        return alignedNCNonROIImage
    else:
        # Return the negative control array
        return arrBaseImage



def applyNegativeControl(baseImage: Union[Image, ndarray],
                         negativeControlType: str = "shuffled",
                         negativeControlRegion: str = "full",
                         roiMask: Optional[Union[Image, ndarray]] = None,
                         randomSeed: Optional[int] = None
) -> Union[Image, ndarray]:
    """Function to apply a negative control to a region of interest (ROI) within a sitk.Image or np.ndarray.

    Parameters
    ----------
    baseImage : sitk.Image | np.ndarray
        Image to apply negative control to. Can be a sitk.Image or np.ndarray.
    negativeControlType : {'shuffled', 'randomized', 'randomized_sampled'}, default 'shuffled'
        Name of negative control to apply.
    negativeControlRegion : {'full', 'roi', 'non_roi'}, default 'full'
        Whether to apply the negative control to the entire image, to the ROI, or to the non-ROI pixels.
    randomSeed : int, default None    
        Value to initialize random number generator with. Set for reproducible results.
    
    Returns
    -------
    sitk.Image | np.ndarray
        Image with negative control function applied to all pixel values within the specificed region of interest.

    """
    if negativeControlType not in ["shuffled", "randomized", "randomized_sampled"]:
        raise ValueError("negativeControlType must be one of 'shuffled', 'randomized', or 'randomized_sampled'")
    if negativeControlRegion not in ["full", "roi", "non_roi"]:
        raise ValueError("regionOfInterest must be one of 'full', 'roi', or 'non_roi'")
    
    if negativeControlRegion == "full":
        if negativeControlType == "shuffled":
            return makeShuffleImage(baseImage, randomSeed)
        elif negativeControlType == "randomized":
            return makeRandomImage(baseImage, randomSeed)
        elif negativeControlType == "randomized_sampled":
            return makeRandomSampleFromDistributionImage(baseImage, randomSeed)
    
    assert roiMask is not None, \
        f"ROI mask is None. Must pass ROI mask to negative control function for {negativeControlType} negative control."
    
    if negativeControlRegion == "roi":
        return negativeControlROIOnly(baseImage, roiMask, negativeControlType, randomSeed)
    else: # negativeControlRegion == "non_roi":
        return negativeControlNonROIOnly(baseImage, roiMask, negativeControlType, randomSeed)

#################################################################
####################### OLD FUNCTIONS ###########################
#################################################################
# def makeRandomRoi(
#     baseImage: sitk.Image, 
#     baseROI: sitk.Image, 
#     roiLabel: Optional[int] = None,
#     randomSeed: Optional[int] = None
# ) -> sitk.Image:
#     """Function to generate random pixel values within the Region of Interest based on the range of values in a sitk Image

#     Parameters
#     ----------
#     baseImage : sitk.Image
#         Image to randomly generate pixel values in Region of Interest
#     baseROI : sitk.Image
#         Image detailing Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.
        
#     Returns
#     -------
#     sitk.Image
#         Image with all pixel values within the Region of Interest randomly generated with same dimensions as input image
#     """
#     # Check the ROI Label exists, if not extract it manually
#     if not roiLabel:
#         roiLabel = getROIVoxelLabel(baseROI)

#     # Initialize variables to track the highest and lowest pixel values in the ROI
#     maxVoxelVal = float("-inf")
#     minVoxelVal = float("inf")

#     # Iterate through baseROI to find the highest and lowest values in baseImage's ROI
#     baseROISize = baseROI.GetSize()
#     for x in range(baseROISize[0]):
#         for y in range(baseROISize[1]):
#             for z in range(baseROISize[2]):
#                 # Ensure only pixels that match the roiLabel of hte image are looked at
#                 if baseROI.GetPixel(x, y, z) == roiLabel:
#                     # Once a pixel is confirmed to be in the ROI get it's corresponding value from the baseImage
#                     current_value = baseImage.GetPixel(x, y, z)
#                     # Update the max and min values
#                     maxVoxelVal = max(maxVoxelVal, current_value)
#                     minVoxelVal = min(minVoxelVal, current_value)

#     # Create a new base image so we are not directly editing the input image
#     new_base = baseImage.__copy__()
#     # Delete the input image to save space
#     del baseImage

#     # Set the random seed for np random generator
#     randNumGen = np.random.default_rng(seed=randomSeed)

#     # Now iterate over the pixels of the ROI in the image and randomly generate a new value for them
#     for x in range(baseROISize[0]):
#         for y in range(baseROISize[1]):
#             for z in range(baseROISize[2]):
#                 if baseROI.GetPixel(x, y, z) == roiLabel:
#                     # Randomly assigning the current value to the range [maxVoxelVal, maxVoxelVal]
#                     mapped_value = int(randNumGen.integers(low=minVoxelVal, high=maxVoxelVal, endpoint=True))

#                     # Set the new pixel value
#                     new_base.SetPixel(x, y, z, mapped_value)

#     return new_base


# def shuffleROI(
#     baseImage: sitk.Image, 
#     baseROI: sitk.Image, 
#     roiLabel: Optional[int] = None,
#     randomSeed: Optional[int] = None,
# ) -> sitk.Image:
#     """Function to shuffle all pixel values within the Region of Interest in a sitk Image

#     Parameters
#     ----------
#     baseImage : sitk.Image
#         Image to shuffle the Region of Interest pixels in
#     baseROI : sitk.Image
#         Image detailing Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.
        
#     Returns
#     -------
#     sitk.Image
#         Image with all pixel values in the Region of Interest randomly shuffled with same dimensions as input image
#     """
#     # Check the ROI Label exists, if not extract it manually
#     if not roiLabel:
#         roiLabel = getROIVoxelLabel(baseROI)

#     # A collection of corresponding value in the BaseImage of all the pixels in the ROI
#     count = []
#     # Iterate through baseROI to store the corresponding value in the baseImage of all the pixels in the ROI
#     baseROISize = baseROI.GetSize()
#     for x in range(baseROISize[0]):
#         for y in range(baseROISize[1]):
#             for z in range(baseROISize[2]):
#                 if baseROI.GetPixel(x, y, z) == roiLabel:
#                     # append the ROI corresponding pixel values to the list
#                     count.append(baseImage.GetPixel(x, y, z))

#     # Create a new base image so we are not directly editing the input image
#     new_base = baseImage.__copy__()
#     # Delete the input image to save space
#     del baseImage

#     # Initialize the random number generator
#     random.seed(randomSeed)
#     # Randomly shuffling the pixel values
#     random.shuffle(count)

#     for x in range(baseROISize[0]):
#         for y in range(baseROISize[1]):
#             for z in range(baseROISize[2]):
#                 if baseROI.GetPixel(x, y, z) == roiLabel:
#                     # Set the value of a pixel in the ROI to be a shuffled value
#                     new_base.SetPixel(x, y, z, count.pop())

#     return new_base


# def makeRandomNonRoi(
#     baseImage: sitk.Image, 
#     baseROI: sitk.Image,
#     roiLabel: Optional[int] = None,
#     randomSeed: Optional[int] = None
# ) -> sitk.Image:
#     """Function to generate random pixel values outside the Region of Interest based on the range of values in a sitk Image

#     Parameters
#     ----------
#     baseImage : sitk.Image
#         Image to randomly generate pixel values in outside the Region of Interest
#     baseROI : sitk.Image
#         Image detailing the Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.

#     Returns
#     -------
#     sitk.Image
#         Image with all pixel values outside the Region of Interest randomly generated with same dimensions as input image
#     """
#     # Check the ROI Label exists, if not extract it manually
#     if not roiLabel:
#         roiLabel = getROIVoxelLabel(baseROI)

#     # Initialize variables to track the highest and lowest pixel values not in the ROI
#     maxVoxelVal: float = float("-inf")
#     minVoxelVal: float = float("inf")

#     # Iterate through baseImage to find the highest and lowest values not in baseImage's ROI
#     baseImageSize = baseImage.GetSize()
#     baseROISize = baseROI.GetSize()
#     for x in range(baseImageSize[0]):
#         for y in range(baseImageSize[1]):
#             for z in range(baseImageSize[2]):
#                 # Ensure only pixels that do not match the roiLabel of the image  or are not in the ROI are looked at
#                 if (
#                     x > baseROISize[0]
#                     or y > baseROISize[1]
#                     or z > baseROISize[2]
#                     or baseROI.GetPixel(x, y, z) != roiLabel
#                 ):
#                     # Once a pixel is confirmed to be outside the ROI, we get it's corresponding value from the baseImage
#                     current_value = baseImage.GetPixel(x, y, z)
#                     # Update the max and min values
#                     maxVoxelVal = max(maxVoxelVal, current_value)
#                     minVoxelVal = min(minVoxelVal, current_value)

#     # Create a new base image so we are not directly editing the input image
#     new_base = baseImage.__copy__()
#     # Delete the input image to save space
#     del baseImage

#     # Set the random seed for np random number generator
#     randNumGen = np.random.default_rng(seed=randomSeed)

#     # Now iterate over the pixels outside the ROI in the image and randomly generate a new value for them
#     for x in range(baseImageSize[0]):
#         for y in range(baseImageSize[1]):
#             for z in range(baseImageSize[2]):
#                 if (
#                     x > baseROISize[0]
#                     or y > baseROISize[1]
#                     or z > baseROISize[2]
#                     or baseROI.GetPixel(x, y, z) != roiLabel
#                 ):
#                     # Randomly assigning the current value to the range [maxVoxelVal, maxVoxelVal]
#                     mapped_value = int(randNumGen.integers(low=minVoxelVal, high=maxVoxelVal, endpoint=True))

#                     # Set the new pixel value
#                     new_base.SetPixel(x, y, z, mapped_value)

#     return new_base


# def shuffleNonROI(
#     baseImage: sitk.Image, 
#     baseROI: sitk.Image,
#     roiLabel: Optional[int] = None,
#     randomSeed: Optional[int] = None,
# ) -> sitk.Image:
#     """Function to shuffle all pixel values that are not within the Region of Interest in a sitk Image

#     Parameters
#     ----------
#     baseImage : sitk.Image
#         Image to shuffle the pixels outside the Region of Interest in
#     baseROI : sitk.Image
#         Image detailing Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.
#     Returns
#     -------
#     sitk.Image
#         Image with all pixel values outside the Region of Interest randomly shuffled with same dimensions as input image
#     """
#     # Check the ROI Label exists, if not extract it manually
#     if not roiLabel:
#         roiLabel = getROIVoxelLabel(baseROI)

#     # A collection of corresponding value in the BaseImage of all the pixels not in the ROI
#     count = []
#     # Iterate through baseImage to store the corresponding value in the baseImage of all the pixels outside the ROI
#     baseImageSize = baseImage.GetSize()
#     baseROISize = baseROI.GetSize()
#     for x in range(baseImageSize[0]):
#         for y in range(baseImageSize[1]):
#             for z in range(baseImageSize[2]):
#                 if (
#                     x > baseROISize[0]
#                     or y > baseROISize[1]
#                     or z > baseROISize[2]
#                     or baseROI.GetPixel(x, y, z) != roiLabel
#                 ):
#                     # Here the key is the pixel coordinate and the value is the value of the pixel in the base image
#                     count.append(baseImage.GetPixel(x, y, z))

#     # Create a new base image so we are not directly editing the input image
#     new_base = baseImage.__copy__()
#     # Delete the input image to save space
#     del baseImage

#     # Initialize the random number generator
#     random.seed(randomSeed)
#     # Randomly shuffling the pixel values
#     random.shuffle(count)

#     for x in range(baseImageSize[0]):
#         for y in range(baseImageSize[1]):
#             for z in range(baseImageSize[2]):
#                 if (
#                     x > baseROISize[0]
#                     or y > baseROISize[1]
#                     or z > baseROISize[2]
#                     or baseROI.GetPixel(x, y, z) != roiLabel
#                 ):
#                     # Set the value of a pixel outside the ROI to be a shuffled value
#                     new_base.SetPixel(x, y, z, count.pop())

#     return new_base





# def makeRandomFromRoiDistribution(
#     baseImage: sitk.Image, 
#     baseROI: sitk.Image, 
#     roiLabel: Optional[int] = None,
#     randomSeed: Optional[int] = None
# ) -> sitk.Image:
#     """Function to randomly sample pixel values within the Region of Interest uniformly from the distribution of pixel values in the ROI region sitk Image

#     Parameters
#     ----------
#     baseImage : sitk.Image
#         Image to randomly generate pixel values in Region of Interest
#     baseROI : sitk.Image
#         Image detailing Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.
        
#     Returns
#     -------
#     sitk.Image
#         Image with all pixel values within the Region of Interest randomly sampled with same dimensions as input image
#     """
#     # Check the ROI Label exists, if not extract it manually
#     if not roiLabel:
#         roiLabel = getROIVoxelLabel(baseROI)

#     # Initialize array of ROI pixel distribution
#     distributionROI = []

#     # Iterate through baseROI to find the highest and lowest values in baseImage's ROI
#     baseROISize = baseROI.GetSize()
#     for x in range(baseROISize[0]):
#         for y in range(baseROISize[1]):
#             for z in range(baseROISize[2]):
#                 # Ensure only pixels that match the roiLabel of hte image are looked at
#                 if baseROI.GetPixel(x, y, z) == roiLabel:
#                     # Once a pixel is confirmed to be in the ROI get it's corresponding value from the baseImage
#                     distributionROI.append(baseImage.GetPixel(x, y, z))

#     # Create a new base image so we are not directly editing the input image
#     new_base = baseImage.__copy__()
#     # Delete the input image to save space
#     del (baseImage)

#     # Set the random seed for np random number generator
#     randNumGen = np.random.default_rng(seed=randomSeed)

#     # Now iterate over the pixels of the ROI in the image and randomly generate a new value for them
#     for x in range(baseROISize[0]):
#         for y in range(baseROISize[1]):
#             for z in range(baseROISize[2]):
#                 if baseROI.GetPixel(x, y, z) == roiLabel:
#                     # Assigning the current value to the randomly sampled value from within the ROI
#                     mapped_value = int(randNumGen.choice(distributionROI))

#                     # Set the new pixel value
#                     new_base.SetPixel(x, y, z, mapped_value)

#     return new_base


# def makeRandomNonRoiFromDistribution(
#     baseImage: sitk.Image, 
#     baseROI: sitk.Image, 
#     roiLabel: Optional[int] = None, 
#     randomSeed: Optional[int] = None
# ) -> sitk.Image:
#     """Function to random sample pixel values outside the Region of Interest uniformly from the distribution of pixel values outside the ROI in a sitk Image

#     Parameters
#     ----------
#     baseImage : sitk.Image
#         Image to randomly generate pixel values in outside the Region of Interest
#     baseROI : sitk.Image
#         Image detailing the Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.

#     Returns
#     -------
#     sitk.Image
#         Image with all pixel values outside the Region of Interest randomly sample form outside the ROI with same dimensions as input image
#     """
#     # Check the ROI Label exists, if not extract it manually
#     if not roiLabel:
#         roiLabel = getROIVoxelLabel(baseROI)

#     # Initialize array of non-ROI pixel distribution
#     distributionROI = []

#     # Iterate through baseImage to find the highest and lowest values not in baseImage's ROI
#     baseImageSize = baseImage.GetSize()
#     baseROISize = baseROI.GetSize()
#     for x in range(baseImageSize[0]):
#         for y in range(baseImageSize[1]):
#             for z in range(baseImageSize[2]):
#                 # Ensure only pixels that do not match the roiLabel of the image  or are not in the ROI are looked at
#                 if (
#                     x > baseROISize[0]
#                     or y > baseROISize[1]
#                     or z > baseROISize[2]
#                     or baseROI.GetPixel(x, y, z) != roiLabel
#                 ):
#                     # Once a pixel is confirmed to be outside the ROI, we get it's corresponding value from the baseImage
#                     distributionROI.append(baseImage.GetPixel(x, y, z))

#     # Create a new base image so we are not directly editing the input image
#     new_base = baseImage.__copy__()
#     # Delete the input image to save space
#     del baseImage

#     # Set the random seed for np random number generator
#     randNumGen = np.random.default_rng(seed=randomSeed)

#     # Now iterate over the pixels outside the ROI in the image and randomly generate a new value for them
#     for x in range(baseImageSize[0]):
#         for y in range(baseImageSize[1]):
#             for z in range(baseImageSize[2]):
#                 if (
#                     x > baseROISize[0]
#                     or y > baseROISize[1]
#                     or z > baseROISize[2]
#                     or baseROI.GetPixel(x, y, z) != roiLabel
#                 ):
#                     # Assigning the current value to the randomly sampled value from within the ROI
#                     mapped_value = int(randNumGen.choice(distributionROI))

#                     # Set the new pixel value
#                     new_base.SetPixel(x, y, z, mapped_value)

#     return new_base


# def applyNegativeControl(
#     nc_type: str,
#     baseImage: sitk.Image,
#     baseROI: Optional[sitk.Image] = None,
#     roiLabel: Optional[int] = None,
#     randomSeed: Optional[int] = None
# ) -> sitk.Image:
#     """Function to generate random pixel values within the Region of Interest based on the range of values in a sitk Image

#     Parameters
#     ----------
#     nc_type : str
#         The type of negative control to be applied
#     baseImage : sitk.Image
#         The image to be modified
#     baseROI : sitk.Image
#         Image detailing Region of Interest
#     roiLabel : int
#         The label representing the ROI in baseROI
#     randomSeed : int
#         Value to initialize random number generator with for shuffling. Set for reproducible results.
        
#     Returns
#     -------
#     sitk.Image
#         The output image with the negative control applied
        
#     Raises
#     ------
#     ValueError
#         If the nc_type is not a valid negative control type
#     """
    
#     if nc_type == "randomized_full":
#         # Make negative control version of ctImage (randomized pixel size)
#         return makeRandomImage(baseImage, randomSeed)
#     elif nc_type == "shuffled_full":
#         # Make negative control version of ctImage (random shuffled pixels, same size)
#         return shuffleImage(baseImage, randomSeed)
#     elif nc_type == "randomized_sampled_full":
#         # Make negative control version of ctImage (random sampled pixels from original distribution, same size)
#         return randomizeImageFromDistributionSampling(baseImage, randomSeed)
    
#     # typesafety check here to ensure baseROI is not None for the following negative control types
#     assert baseROI is not None, \
#         f"baseROI must be provided for {nc_type} negative control type"
    
#     if nc_type == "randomized_roi":
#         # Make negative control version of ctImage (randomized pixel size inside the ROI)
#         return makeRandomRoi(baseImage, baseROI, roiLabel, randomSeed)
#     elif nc_type == "shuffled_roi":
#         # Make negative control version of ctImage (random shuffled pixels inside the ROI, same size)
#         return shuffleROI(baseImage, baseROI, roiLabel, randomSeed)
#     elif nc_type == "randomized_non_roi":
#         # Make negative control version of ctImage (randomized pixel size outside the ROI)
#         return makeRandomNonRoi(baseImage, baseROI, roiLabel, randomSeed)
#     elif nc_type == "shuffled_non_roi":
#         # Make negative control version of ctImage (shuffled pixels outside the ROI, same size)
#         return shuffleNonROI(baseImage, baseROI, roiLabel, randomSeed)
#     elif nc_type == "randomized_sampled_roi":
#         # Make negative control version of ctImage (random sampled pixels from original distribution inside ROI, same size)
#         return makeRandomFromRoiDistribution(baseImage, baseROI, roiLabel, randomSeed)
#     elif nc_type == "randomized_sampled_non_roi":
#         # Make negative control version of ctImage (random sampled pixels from original distribution outside ROI, same size)
#         return makeRandomNonRoiFromDistribution(baseImage, baseROI, roiLabel, randomSeed)
#     else:
#         raise ValueError("Invalid nc_type. Please choose a valid negative control type.")



"""Module for loading different data types for the READII pipeline."""

from .features import loadFeatureFilesFromImageTypes
from .general import loadFileToDataFrame, loadImageDatasetConfig
from .images import getImageTypesFromDirectory

__all__ = [
    "loadFeatureFilesFromImageTypes",
    "loadFileToDataFrame",
    "loadImageDatasetConfig",
    "getImageTypesFromDirectory"
]


from pathlib import Path
from typing import Union

from readii.utils import logger


def getImageTypesFromDirectory(raw_data_dir:Union[Path|str],
                               feature_file_prefix:str = "",
                               feature_file_suffix:str = ".csv") -> list:
    """Get a list of image types from a directory containing image feature files.

    Parameters
    ----------
    raw_data_dir : str
        Path to the directory containing the image feature files.
    feature_file_prefix : str, optional
        Prefix to remove from the feature file name. The default is "".
    feature_file_suffix : str, optional
        Suffix to remove from the feature file name. The default is ".csv".
    
    Returns
    -------
    list
        List of image types from the image feature files.
    """
    # Check if raw_data_dir is a string or a Path object, convert to Path object if it is a string
    if isinstance(raw_data_dir, str):
        raw_data_dir = Path(raw_data_dir)

    # Check if the directory exists
    if not raw_data_dir.exists():
        msg = f"Directory {raw_data_dir} does not exist."
        logger.error(msg)
        raise FileNotFoundError()
    
    # Check if the directory is a directory
    if not raw_data_dir.is_dir():
        msg = f"Path {raw_data_dir} is not a directory."
        logger.error(msg)
        raise NotADirectoryError()
    
    # Check that directory contains files with the specified prefix and suffix
    if not any(raw_data_dir.glob(f"{feature_file_prefix}*{feature_file_suffix}")):
        msg = f"No files with prefix {feature_file_prefix} and suffix {feature_file_suffix} found in directory {raw_data_dir}."
        logger.error(msg)
        raise FileNotFoundError()

    # Initialize an empty list to store the image types
    image_types = []

    # Get list of file banes with the specified prefix and suffix in the directory
    for file in raw_data_dir.glob(f"{feature_file_prefix}*{feature_file_suffix}"):
        file_name = file.name

        # Remove the prefix and suffix from the file name
        image_type = file_name.removeprefix(feature_file_prefix).removesuffix(feature_file_suffix)
        
        # Add the image type to the list
        image_types.append(image_type)

    return image_types


from argparse import ArgumentParser
import os

from readii.metadata import *
from readii.feature_extraction import *

from readii.utils import logger

def parser():
    """Function to take command-line arguments and set them up for the pipeline run
    """
    parser = ArgumentParser("READII Feature Extraction Pipeline")

    # arguments
    parser.add_argument("data_directory", type=str,
                        help="Path to top-level directory of image dataset. Same as med-imagetools.")
    
    parser.add_argument("output_directory", type=str,
                       help="Path to output directory to save radiomic features and metadata.")
    
    parser.add_argument("--roi_names", type=str, default=None,
                        help="Name of region of interest in RTSTRUCT to perform extraction on.")
    
    parser.add_argument("--pyradiomics_setting", type=str, default=None,
                        help="Path to PyRadiomics configuration YAML file. If none provided, will use \
                              default in src/readii/data/.")
    
    parser.add_argument("--negative_controls", type=str, default=None,
                        help="List of negative control types to run feature extraction on. Input as comma-separated list with no spaces.  \
                              Options: randomized_full,randomized_roi,randomized_non_roi,shuffled_full,shuffled_roi,shuffled_non_roi,randomized_sampled_full,randomized_sampled_roi,randomized_sampled_non_roi")

    parser.add_argument("--parallel", action="store_true",
                        help="Whether to run feature extraction in a parallel process. False by default.")

    parser.add_argument("--update", action="store_true", help="Flag to force rerun all steps of pipeline. False by default.")

    parser.add_argument("--random_seed", type=int,
                        help="Value to set random seed to for reproducible negative controls")

    parser.add_argument("--keep_running", action="store_true",
                        help="Flag to keep pipeline running even when feature extraction for a patient fails. False by default.")

    return parser.parse_known_args()[0]
    

    

def main():
    """Function to run READII radiomic feature extraction pipeline.
    """
    args = parser()
    pretty_args = '\n\t'.join([f"{k}: {v}" for k, v in vars(args).items()])
    logger.debug(
        f"Arguments:\n\t{pretty_args}"
    )

    logger.info("Starting readii pipeline...", args=vars(args))

    # Set up output directory
    outputDir = os.path.join(args.output_directory, "readii_outputs")
    if not os.path.exists(outputDir):
        logger.info(f"Directory {outputDir} does not exist. Creating...")
        os.makedirs(outputDir)
    else:
        logger.warning(f"Directory {outputDir} already exists. Will overwrite contents.")

    # Find med-imagetools output files
    logger.info("Finding med-imagetools outputs...")
    parentDirPath, datasetName = os.path.split(args.data_directory)
    imageFileListPath = os.path.join(parentDirPath + "/.imgtools/imgtools_" + datasetName + ".csv")
    if not os.path.exists(imageFileListPath):
        # Can we run med-imagetools in here?
        logger.error(
            f"Expected file {imageFileListPath} not found. Check the data_directory argument or run med-imagetools."
        )
        raise FileNotFoundError("Output for med-imagetools not found for this image set. Check the data_directory argument or run med-imagetools.")

    logger.info("Getting segmentation type...")
    try:
        # Get segType from imageFileList to generate the image metadata file and set up feature extraction
        segType = getSegmentationType(imageFileListPath)
    except RuntimeError as e:
        logger.error(str(e))
        logger.error("Feature extraction not complete.")
        exit()


    # Check if image metadata file has already been created
    imageMetadataPath = createImageMetadataFile(
        outputDir, 
        parentDirPath, 
        datasetName, 
        segType, 
        imageFileListPath, 
        args.update)
    
    # Check if radiomic feature file already exists
    radFeatOutPath = os.path.join(outputDir, "features/", "radiomicfeatures_original_" + datasetName + ".csv")
    if not os.path.exists(radFeatOutPath) or args.update:
        logger.info("Starting radiomic feature extraction...")
        radiomicFeatures = radiomicFeatureExtraction(imageMetadataPath = imageMetadataPath,
                                                     imageDirPath = parentDirPath,
                                                     roiNames = args.roi_names,
                                                     pyradiomicsParamFilePath = args.pyradiomics_setting,
                                                     outputDirPath = outputDir,
                                                     negativeControl = None,
                                                     parallel = args.parallel,
                                                     keep_running = args.keep_running)
    else:
        logger.info(f"Radiomic features have already been extracted. See {radFeatOutPath}")

    # Negative control radiomic feature extraction
    if args.negative_controls != None:
        # Get all negative controls to run
        negativeControlList = args.negative_controls.split(",")

        # Perform feature extraction for each negative control type
        for negativeControl in negativeControlList:
            ncRadFeatOutPath = os.path.join(outputDir, "features/", "radiomicfeatures_" + negativeControl + "_" + datasetName + ".csv")
            if not os.path.exists(ncRadFeatOutPath) or args.update:
                logger.info(f"Starting radiomic feature extraction for negative control: {negativeControl}")
                ncRadiomicFeatures = radiomicFeatureExtraction(imageMetadataPath = imageMetadataPath,
                                                               imageDirPath = parentDirPath,
                                                               roiNames = args.roi_names,
                                                               pyradiomicsParamFilePath = args.pyradiomics_setting,
                                                               outputDirPath = outputDir,
                                                               negativeControl = negativeControl,
                                                               randomSeed=args.random_seed,
                                                               parallel = args.parallel,
                                                               keep_running = args.keep_running)
            else:
                logger.info(f"{negativeControl} radiomic features have already been extracted. See {ncRadFeatOutPath}")

    
    logger.info("Pipeline complete.")

if __name__ == "__main__":
    main()


from pathlib import Path
from typing import Dict, Union

from readii.io.loaders.general import loadImageDatasetConfig
from readii.utils import logger


def get_full_data_name(config: Union[Dict | Path]) -> str:
    """Combine DATA_SOURCE and DATASET_NAME config variables into a single string."""
    match type(config):
        case Path():
            dataset_config = loadImageDatasetConfig(config.stem, config.parent)
        case Dict():
            dataset_config = config
        case _:
            message = "Error getting full data name: config must be of type Path or Dict."
            logger.debug(message)
            raise TypeError(message) 

    try:
        data_source = dataset_config['DATA_SOURCE']
        dataset_name = dataset_config['DATASET_NAME']
        return f"{data_source}_{dataset_name}"
    except KeyError as e:
        message = f"Missing required key in dataset configuration: {e}"
        logger.error(message)
        raise ValueError(message) from e